/*
 *	rdate --- remote date, adjust clock referring clock on remote host.
 *
 * Copyright (c) 2003
 * 	Koh-ichi Ito. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the developer nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif	/* HAVE_CONFIG_H */
#include <sys/types.h>
#include <netinet/in.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/uio.h>
#include <sys/param.h>
#include <netdb.h>
#include <time.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif	/* HAVE_UNISTD_H */

#define MODE_AUTO	0
#define MODE_JUMP	1
#ifdef HAVE_ADJTIME
#define OPTSTRING	"bl:uv"
#else
#define OPTSTRING	"buv"
#endif
#ifndef MAX_DIFF
#define MAX_DIFF	60L
#endif
/*
 *	"Time Protocol"(RFC868,	ftp://ftp.rfc-editor.org/in-notes/rfc868.txt)
 *	returns the seconds since 00:00:00 1 Jan 1900 GMT.
 *	On the other hand, time(3) and the family in libc handle
 *	the seconds since 00:00:00 1 Jan 1970 UTC, what we call 'Epoch'.
 *
 *	MAGIC_1970 is the difference between them ignoreing the difference
 *	between GMT and UTC. The number is mentioned in RFC868 as an example.
 */
#define SERVICE	"time"
#define MAGIC_1970	2208988800UL

int	tcp_doit();
int	trim();
void	usage();

static char	Id[] = "$Id: rdate.c,v 1.1.1.1 2003/11/26 08:39:54 kohi Exp $";
char	*MyName;
int	Verbose = 0;

main(argc, argv)

	int	argc;
	char	**argv;
{
	int	optch,
		mode,
		max_diff;

	MyName = *argv;
#ifdef HAVE_ADJTIME
	mode = MODE_AUTO;
#else
	mode = MODE_JUMP;	/* never be MODE_AUTO */
#endif	/* HAVE_ADJTIME */
	max_diff = MAX_DIFF;
	while ( (optch = getopt(argc, argv, OPTSTRING) ) != -1) {
		switch (optch) {
		case 'b':
			mode = MODE_JUMP;
			break;
#ifdef HAVE_ADJTIME
		case 'l':
			max_diff = atoi(optarg);
			break;
#endif	/* HAVE_ADJTIME */
		case 'u':
			fprintf(stderr, "%s: sorry, udp routine is not implemented yet.\n");
			exit(1);
		case 'v':
			Verbose = 1;
			break;
		default:
			usage();
			/* never comin' back.*/
		}
	}
	argc -= optind;
	argv += optind;
	if (argc < 1) {
		usage();
		/* never comin' back.*/
	}
	if (!tcp_doit(*argv, mode, max_diff) ) {
		err(1, "tcp_doit()");
		/* never comin' back.*/
	}
	exit(0);
}

int tcp_doit(hostname, mode, max_diff)

	char	*hostname;
	int	mode,
		max_diff;
{
	int	s,
		gai_err;
	time_t	remote_time,
		t0,
		t1;
	struct addrinfo	hints,
			*res,
			*ai;

	memset(&hints, 0, sizeof(hints) );
	hints.ai_family = AF_UNSPEC;
	hints.ai_socktype = SOCK_STREAM;
	hints.ai_protocol = IPPROTO_TCP;
	if ( (gai_err = getaddrinfo(hostname,
			SERVICE, &hints, &res) ) != 0) {
		fprintf(stderr, "getaddrinfo(%s) failed: %s\n",
			hostname, gai_strerror(gai_err) );
		return 0;
	}
	for (ai = res; ai; ai = ai->ai_next) {
		if ( (s = socket(ai->ai_family, ai->ai_socktype,
				ai->ai_protocol) ) < 0) {
			continue;
		}
		if (connect(s, ai->ai_addr, ai->ai_addrlen) == 0) {
			break;
		}
		close(s);
	}
	if (res) {
		freeaddrinfo(res);
	}
	if (!ai) {
		return 0;
	}
	t0 = time(NULL);
	if (read(s, &remote_time, sizeof(time_t) ) < sizeof(time_t) ) {
		perror("read()");
		close(s);
		return 0;
	}
	t1 = time(NULL);
	close(s);
	/* guess the latency consumed by the return path.*/
	remote_time = ntohl(remote_time) - MAGIC_1970 + (t1 - t0) / 2;
	if (Verbose) {
		printf("local:  %s", ctime(&t1) );
		printf("remote: %s", ctime(&remote_time) );
		if (t1 - t0) {
			printf("round trip takes %d sec.\n", (t1 - t0) );
		}
	}
	return trim(t1, remote_time, mode, max_diff);
}

int trim(tlocal, tremote, mode, max_diff)

	time_t	tlocal,
		tremote;
	int	mode,
		max_diff;
{
	struct timeval	tv;

#ifdef HAVE_ADJTIME
	if ( (mode == MODE_JUMP)
	  || (max_diff < abs( (int)(tlocal - tremote) ) ) ) {
#endif	/* HAVE_ADJTIME */
		if (Verbose) {
			printf("jump mode\n");
		}
		tv.tv_sec = (long)tremote;
		tv.tv_usec = 0L;	/* Don't expect so much.*/
		if (settimeofday(&tv, NULL) != 0) {
			warn("settimeofday()");
			return 0;
		}
#ifdef HAVE_ADJTIME
	} else {
		if (Verbose) {
			printf("slow mode\n");
		}
		tv.tv_sec = (long)(tremote - tlocal);
		tv.tv_usec = 0L;	/* Don't expect so much.*/
		if (adjtime(&tv, NULL) != 0) {
			warn("adjtime():");
			return 0;
		}
	}
#endif	/* HAVE_ADJTIME */
	return 1;
}

void usage()
{
	fprintf(stderr,
#ifdef HAVE_ADJTIME
	"Usage: %s [-bv][-l limit] hostname\n",
#else
	"Usage: %s [-bv] hostname\n",
#endif
	MyName);
	exit(1);
}
